"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.dict = exports.kv = exports.KV = exports.Dictionary = void 0;
const hash_1 = require("../hash");
const equality_1 = require("../equality");
const set_1 = require("../set/set");
const sequence_1 = require("../sequence");
const optional_1 = require("../optional");
const codable_1 = require("../codable");
const clone_1 = require("../clone");
class Dictionary {
    constructor(hashKV) {
        this.hashKV = hashKV;
    }
    /** 引数のキーに対応する値を返す */
    get(key) {
        const hashValue = (0, hash_1.hash)(key);
        const kvSetOpt = (0, optional_1.optional)(this.hashKV.get(hashValue));
        const kvOpt = kvSetOpt.flatMap((kvs) => kvs.find((kv) => (0, equality_1.equals)(kv.key, key)));
        return kvOpt.map((kv) => kv.value);
    }
    /** 対応するキーの値を更新する。none系を渡すとキーが削除される */
    put(key, value) {
        var _a;
        const hashValue = (0, hash_1.hash)(key);
        const copy = this.shallowCopyHashKV;
        const oldKvSet = (_a = copy.get(hashValue)) !== null && _a !== void 0 ? _a : (0, set_1.set)();
        const newKvSet = (0, optional_1.optional)(value).fold(() => oldKvSet.filter((kv) => !(0, equality_1.equals)(kv.key, key)), (v) => oldKvSet.insert(kv(key, v)));
        return new Dictionary(copy.set(hashValue, newKvSet));
    }
    get shallowCopyHashKV() {
        const hashKV = new Map();
        for (const [hash, kvs] of this.hashKV) {
            hashKV.set(hash, kvs.clone);
        }
        return hashKV;
    }
    equals(other) {
        return this.allKvSet.equals(other.allKvSet);
    }
    get hashValue() {
        return (0, hash_1.hash)("dict" + this.allKvSet.hashValue);
    }
    map(f) {
        const result = this.allKvSeq.map((kv) => f(kv.key, kv.value));
        let next = dict();
        for (const [key, value] of result) {
            next = next.put(key, value);
        }
        return next;
    }
    filter(f) {
        const result = this.allKvSeq.filter((kv) => f(kv.key, kv.value));
        let next = dict();
        for (const kv of result) {
            next = next.put(kv.key, kv.value);
        }
        return next;
    }
    every(f) {
        return this.allKvSeq.every((kv) => f(kv.key, kv.value));
    }
    some(f) {
        return this.allKvSeq.some((kv) => f(kv.key, kv.value));
    }
    get encode() {
        return this.allKvSeq.reduce({}, (record, kv) => {
            const encodedKey = (0, codable_1.encode)(kv.key);
            if (typeof encodedKey === "string" || typeof encodedKey === "number") {
                record[encodedKey] = (0, codable_1.encode)(kv.value);
                return record;
            }
            return record;
        });
    }
    decode(object) {
        if (typeof object !== "object" || object == null) {
            throw new Error(`${object} should be object`);
        }
        if (Array.isArray(object)) {
            throw new Error(`${object} should be object, but it is array`);
        }
        if (this.isEmpty) {
            throw new Error(`${this} is empty Dictionary`);
        }
        const { key, value } = this.allKvSeq.get(0);
        if (!(typeof key === "string" || typeof key === "number" || (0, codable_1.isDecodable)(key))) {
            throw new Error("Dictionary key should be string or number or Decodable");
        }
        let dictionary = dict();
        for (const jsonKey in object) {
            dictionary = dictionary.put(jsonKey, (0, codable_1.decode)(object[jsonKey], value));
        }
        return dictionary;
    }
    get clone() {
        let result = dict();
        for (const kv of this.allKvSeq) {
            result = result.put((0, clone_1.clone)(kv.key), (0, clone_1.clone)(kv.value));
        }
        return result;
    }
    get isEmpty() {
        return this.count === 0;
    }
    get nonEmpty() {
        return this.count > 0;
    }
    get count() {
        return this.allKvSeq.count;
    }
    get keys() {
        return this.allKvSeq.map((kv) => kv.key);
    }
    get values() {
        return this.allKvSeq.map((kv) => kv.value);
    }
    [Symbol.iterator]() {
        return this.allKvArrays[Symbol.iterator]();
    }
    get allKvSet() {
        return (0, set_1.set)(this.allKvSeq);
    }
    get allKvArrays() {
        return this.allKvSeq.map((kv) => [kv.key, kv.value]).toArray;
    }
    get allKvSeq() {
        let allKvSeq = (0, sequence_1.seq)();
        for (const [_, kvs] of this.hashKV) {
            allKvSeq = allKvSeq.concat(kvs.toSeq);
        }
        return allKvSeq;
    }
}
exports.Dictionary = Dictionary;
class KV extends hash_1.Hashable {
    constructor(key, value) {
        super();
        this.key = key;
        this.value = value;
    }
    get properties() {
        return [this.key];
    }
}
exports.KV = KV;
function kv(key, value) {
    return new KV(key, value);
}
exports.kv = kv;
function dict(array) {
    var _a, _b;
    if (array == null)
        return new Dictionary(new Map());
    if (Array.isArray(array) || array instanceof sequence_1.Sequence) {
        const sequence = (0, sequence_1.seq)(array);
        const hashKV = new Map();
        for (const kv of sequence) {
            const hashValue = kv.hashValue;
            const origin = (_a = hashKV.get(hashValue)) !== null && _a !== void 0 ? _a : (0, set_1.set)();
            hashKV.set(hashValue, origin.insert(kv));
        }
        return new Dictionary(hashKV);
    }
    const hashKV = new Map();
    for (const key in array) {
        const hashValue = (0, hash_1.hash)(key);
        const origin = (_b = hashKV.get(hashValue)) !== null && _b !== void 0 ? _b : (0, set_1.set)();
        const kv = new KV(key, array[key]);
        hashKV.set(hashValue, origin.insert(kv));
    }
    return new Dictionary(hashKV);
}
exports.dict = dict;
//# sourceMappingURL=dictionary.js.map