"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.seq = exports.Sequence = void 0;
const equality_1 = require("../equality");
const clone_1 = require("../clone");
const optional_1 = require("../optional");
const range_1 = require("../range");
const compare_1 = require("../compare");
const codable_1 = require("../codable");
const hash_1 = require("../hash");
class Sequence {
    constructor(values) {
        this.values = values;
    }
    /** 要素が空の場合のみtrueを返す */
    get isEmpty() {
        return this.count === 0;
    }
    /** 要素が空の場合のみfalseを返す */
    get nonEmpty() {
        return !this.isEmpty;
    }
    /** 最初の要素を取得する */
    get first() {
        return (0, optional_1.optional)(this.values[0]);
    }
    /** 最後の要素を取得する */
    get last() {
        return this.at(-1);
    }
    /** 要素を取り出す */
    get(index) {
        return this.values[index];
    }
    /** 要素を取り出す */
    at(index) {
        return (0, optional_1.optional)(this.values.at(index));
    }
    /** 最後尾に値を追加する */
    append(value) {
        return new Sequence([...this.values, value]);
    }
    /** 間に値を挿入する */
    insert(element, at) {
        return new Sequence([
            ...this.values.slice(0, at),
            element,
            ...this.values.slice(at),
        ]);
    }
    remove(at) {
        return new Sequence([
            ...this.values.slice(0, at),
            ...this.values.slice(at + 1),
        ]);
    }
    /** 文字列の配列を結合する */
    join(separator) {
        return this.values.join(separator);
    }
    /** 要素を変換する */
    map(f) {
        return new Sequence(this.values.map((v, i, arr) => f(v, i, seq(arr))));
    }
    /** 要素をフラットに変換する */
    flatMap(f) {
        return new Sequence(this.values.flatMap((v, i, arr) => f(v, i, seq(arr)).toArray));
    }
    /** unwrapしてsomeのみでmapを行う */
    compactMap(f) {
        return this.filter((option) => (0, optional_1.isOptional)(option) && option.isDefined)
            .map((option) => option.get)
            .map((v, i, sequence) => f(v, i, sequence));
    }
    /** 要素をフィルタリングする */
    filter(f) {
        return new Sequence(this.values.filter((v, i, arr) => f(v, i, seq(arr))));
    }
    /** 連結する */
    concat(elements) {
        return new Sequence(this.values.concat(elements.values));
    }
    /** 全ての要素が条件を満たす際にtrueを返す */
    every(f) {
        return this.values.every((value, index, array) => f(value, index, seq(array)));
    }
    /** どれかの要素が条件を満たす際にtrueを返す */
    some(f) {
        return this.values.some((value, index, array) => f(value, index, seq(array)));
    }
    /** ソートする。元のSequenceに影響はない */
    sorted(sorter) {
        const result = this.clone.values.slice().sort((a, b) => {
            if (sorter == null) {
                return (0, compare_1.compareTo)(a, b);
            }
            if ((0, equality_1.equals)(a, b))
                return 0;
            const result = sorter(a, b);
            if (typeof result === "boolean")
                return result ? -1 : 1;
            return result;
        });
        return new Sequence(result);
    }
    /** 最大値を取得する */
    max(sorter) {
        return this.sorted(sorter).last;
    }
    /** 最小値を取得する */
    min(sorter) {
        return this.sorted(sorter).first;
    }
    /** 検索する */
    find(f) {
        return (0, optional_1.optional)(this.values.find((value, index, array) => f(value, index, seq(array))));
    }
    /** index付きのSequenceを返す */
    get enumerated() {
        return this.map((value, index) => ({ index, value }));
    }
    /** 要素を逆順に並び替えたものを返す */
    get reversed() {
        return this.reduce(seq(), (reverse, element) => {
            return reverse.insert(element, 0);
        });
    }
    /** 要素が存在するか判別する */
    contains(element) {
        return this.some((value) => (0, equality_1.equals)(value, element));
    }
    firstIndex(of) {
        return this.enumerated
            .find(({ value }) => (0, equality_1.equals)(value, of))
            .map((value) => value.index);
    }
    /** 要素をランダムに並び替えたものを返す */
    get shuffled() {
        let result = seq();
        let sequence = this.clone;
        while (sequence.nonEmpty) {
            const random = sequence.enumerated.randomElement().get;
            sequence = sequence.remove(random.index);
            result = result.append(random.value);
        }
        return result;
    }
    /** 入れ替えられる */
    swap(a, b) {
        const values = this.values.slice();
        const temp = values[a];
        values[a] = values[b];
        values[b] = temp;
        return new Sequence(values);
    }
    /** 範囲の要素を置き換える */
    replaceSubrange(subrange, withNewElements) {
        const elements = seq(withNewElements);
        if (subrange instanceof range_1.OpenRange) {
            return new Sequence([
                ...this.values.slice(0, subrange.minimum),
                ...elements.values,
                ...this.values.slice(subrange.minimum + elements.count),
            ]);
        }
        if (subrange instanceof range_1.Range) {
            return new Sequence([
                ...this.values.slice(0, subrange.minimum),
                ...elements.values,
                ...this.values.slice(subrange.maximum),
            ]);
        }
        if (subrange instanceof range_1.ClosedRange) {
            return new Sequence([
                ...this.values.slice(0, subrange.minimum),
                ...elements.values,
                ...this.values.slice(subrange.maximum + 1),
            ]);
        }
        return this;
    }
    /** 要素を集約する */
    reduce(init, f) {
        return this.values.reduce(f, init);
    }
    /** 指定した範囲を取得する */
    slice(range) {
        if (range instanceof range_1.OpenRange) {
            return new Sequence(this.values.slice(range.minimum));
        }
        if (range instanceof range_1.Range) {
            return new Sequence(this.values.slice(range.minimum, range.maximum));
        }
        if (range instanceof range_1.ClosedRange) {
            return new Sequence(this.values.slice(range.minimum, range.maximum + 1));
        }
        throw new Error("not Range");
    }
    /** ランダムな要素を取得する。cryptoを使用 */
    randomElement() {
        if (this.isEmpty)
            return (0, optional_1.none)();
        return this.at(Math.floor((Math.random() * 100000) % this.count));
    }
    equals(other) {
        if (this.count !== other.count)
            return false;
        return this.values.every((value, index) => (0, equality_1.equals)(value, other.values[index]));
    }
    get clone() {
        return new Sequence(this.values.map(clone_1.clone));
    }
    /** 要素数を返す */
    get count() {
        return this.values.length;
    }
    get distinct() {
        return this.reduce(seq(), (result, element) => {
            if (result.contains(element))
                return result;
            return result.append(element);
        });
    }
    [Symbol.iterator]() {
        return this.values[Symbol.iterator]();
    }
    next() {
        return this.values[Symbol.iterator]().next();
    }
    get encode() {
        return this.values.map(codable_1.encode);
    }
    decode(object) {
        if (this.isEmpty) {
            throw new Error("empty sequence cannot decode");
        }
        if (!Array.isArray(object)) {
            throw new Error("not array");
        }
        if (!(0, codable_1.isDecodable)(this.get(0))) {
            return seq(object);
        }
        return seq(object).reduce(seq(), (result, element) => {
            return result.append(this.get(0).decode(element));
        });
    }
    get hashValue() {
        return (0, hash_1.hash)(this.values);
    }
    get toArray() {
        return this.values;
    }
}
exports.Sequence = Sequence;
function seq(...elements) {
    if (elements.length === 1) {
        if (Array.isArray(elements[0])) {
            // array: readonly T[]
            return new Sequence(elements[0]);
        }
        if (elements[0] instanceof Sequence) {
            // sequence: Sequence<T>
            return elements[0];
        }
    }
    return new Sequence(elements);
}
exports.seq = seq;
//# sourceMappingURL=sequence.js.map